import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torch.utils.data.dataset import TensorDataset
from .models import SIGRN
from .evaluate import get_metrics_auc
from tqdm import tqdm
from .logger import LightLogger
def runSIGRN(exp_array, configs,
               ground_truth=None, logger=None, progress_bar=False):
    '''
    Initialize and Train a SIGRN model with configs

    Parameters
    ----------
    exp_array: np.array
        Expression data with cells on rows and genes on columns.
    configs: dict
        A dictionary defining various hyperparameters of the
        model. See Hyperparameters include `train_split`,
        `train_split_seed`, `batch_size`, `hidden_dim`, `z_dim`,
        `train_on_non_zero`, `dropout_augmentation`, `cuda`,
        `alpha`, `beta`, `delayed_steps_on_sparse`, `n_epochs`,
        `eval_on_n_steps`, `lr_nn`, `lr_adj`, `K1`, and `K2`.
    ground_truth: tuple or None
        (Optional, only for BEELINE evaluation) You don't need
        to define this parameter when you execute DAZZLE on real
        datasets when the ground truth network is unknown. For
        evaluations on BEELINE,
        BEELINE ground truth object exported by
        data.load_beeline_ground_truth. The first element of this
        tuple is eval_flat_mask, the boolean mask on the flatten
        adjacency matrix to identify TFs and target genes. The
        second element is the lable values y_true after flatten.
    logger: LightLogger or None
        Either a predefined logger or None to start a new one. This
        logger contains metric information logged during training.
    progress_bar: bool
        Whether to display a progress bar on epochs.

    Returns
    -------
    (torch.Module, List)
        This function returns a tuple of the trained model and a list of
        adjacency matrix at all evaluation points.
    '''
    if configs['early_stopping'] != 0 and configs['train_split'] == 1.0:
        raise Exception(
            "You indicate early stopping but you have not specified any ",
            "validation data. Consider decrease your train_split. ")
    es = configs['early_stopping']

    n_obs, n_gene = exp_array.shape

    # Logger -------------------------------------------------------------------
    if logger is None:
        logger = LightLogger()
    logger.set_configs(configs)
    note_id = logger.start()

    # Global Mean/Std ----------------------------------------------------------
    global_mean = torch.FloatTensor(exp_array.mean(0))
    global_std = torch.FloatTensor(exp_array.std(0))

    # Train/Test split if requested --------------------------------------------
    assert configs['train_split'] > 0 and configs['train_split'] <= 1, \
        f'Expect 0<configs["train_split"]<=1'
    has_train_test_split = (configs['train_split'] != 1.0)

    if configs['train_split_seed'] is None:
        train_mask = np.random.rand(n_obs)
    else:
        rs = np.random.RandomState(seed=configs['train_split_seed'])
        train_mask = rs.rand(n_obs)

    train_dt = TensorDataset(
        torch.FloatTensor(exp_array[train_mask <= configs['train_split'],]),
    )
    train_loader = DataLoader(
        train_dt, batch_size=configs['batch_size'], shuffle=True)

    if has_train_test_split:
        val_dt = TensorDataset(
            torch.FloatTensor(exp_array[train_mask > configs['train_split'],]),
        )
        val_loader = DataLoader(
            val_dt, batch_size=configs['batch_size'], shuffle=True)

    # Defining Model
    vae = SIGRN(
        n_gene=n_gene,
        hidden_dim=configs['hidden_dim'], z_dim=configs['z_dim'],
        A_dim=configs['A_dim'],
        train_on_non_zero=configs['train_on_non_zero'],
        dropout_augmentation_p=configs['dropout_augmentation_p'],
        dropout_augmentation_type=configs['dropout_augmentation_type']
        # A_dim=configs['A_dim']
    )
    # Move things to cuda if necessary
    if configs['cuda']:
        global_mean = global_mean.cuda()
        global_std = global_std.cuda()
        vae = vae.cuda()

    if configs['number_of_opt'] == 2:
        opt_nn_e = torch.optim.Adam(vae.inference_zposterior.parameters(), lr=configs['lr_nn'])
        opt_nn_d = torch.optim.Adam(vae.generative_pxz.parameters(), lr=configs['lr_nn'])
        opt_adj = torch.optim.Adam([vae.adj_A], lr=configs['lr_adj'], weight_decay=0.00, betas=[0.9, 0.9])

    es_tracks = []
    adjs = []
    disable_tqdm = (progress_bar == False)
    for epoch in tqdm(range(configs['n_epochs']), disable=disable_tqdm):
        if configs['number_of_opt'] == 2:
            vae.train(True)
            # iteration_for_A = epoch % (configs['K1'] + configs['K2']) >= configs['K1']
            # vae.adj_A.requires_grad = iteration_for_A
            evaluation_turn = (epoch % configs['eval_on_n_steps'] == 0)

            # go through training samples
            eval_log = {
                'epoch': 0, 'loss_rec': 0, 'loss_kl': 0, 'loss_sparse': 0,
                 'loss': 0, 'loss_ma': 0, 'lossG': 0
            }
            for i, batch in enumerate(train_loader):
                exp = batch[0]
                noise = torch.randn_like(exp)
                if configs['cuda']:
                    exp = exp.cuda()
                    noise = noise.cuda()

                # (1)train VAE
                opt_adj.zero_grad()
                opt_nn_e.zero_grad()
                opt_nn_d.zero_grad()

                out_real = vae(exp, global_mean, global_std)

                z_posterior = vae.inference_zposterior(out_real['x_rec'].detach().unsqueeze(-1))
                z_posterior = torch.einsum('ogd,agh->ohd', z_posterior, out_real['IA'])

                z_mu = z_posterior[:, :, :vae.z_dim]
                z_logvar = z_posterior[:, :, vae.z_dim:]
                kl_rec = -0.5 * (1 + z_logvar - z_mu.pow(2) - torch.exp(z_logvar))
                lossE_kl = (- kl_rec).exp().mean()
                lossE_rec_kl = lossE_kl
                lossE_real_kl = out_real['loss_kl']

                loss_margin = lossE_real_kl + lossE_rec_kl

                lossE = loss_margin +  (out_real['loss_rec'])

                adj_m = vae.get_adj_()
                loss_sparse = torch.norm(adj_m, 1) / n_gene / n_gene
                if epoch >= configs['delayed_steps_on_sparse']:
                    lossE += configs['alpha'] * loss_sparse

                lossE.backward(retain_graph=True)


                opt_nn_e.step()
                opt_nn_d.step()

                # (2)updata G and A

                z_posterior_G = vae.inference_zposterior(out_real['x_rec'].detach().unsqueeze(-1))
                z_posterior_G = torch.einsum('ogd,agh->ohd', z_posterior_G, out_real['IA'])

                z_mu_G = z_posterior_G[:, :, :vae.z_dim]
                z_logvar_G = z_posterior_G[:, :, vae.z_dim:]
                lossG_kl = -0.5 * torch.mean(1 + z_logvar_G - z_mu_G.pow(2) - torch.exp(z_logvar_G))

                lossG_rec_kl = lossG_kl

                lossG =  lossG_rec_kl

                lossG = lossG + out_real['loss_rec'].detach()

                lossG.backward()

                opt_adj.step()
                opt_nn_d.step()
                # opt_nn_c.step()

                eval_log['epoch'] += epoch  # lry add
                eval_log['loss_rec'] += round(out_real['loss_rec'].detach().cpu().item(), 5)
                eval_log['loss_kl'] += round(out_real['loss_kl'].detach().cpu().item(), 5)
                eval_log['loss_sparse'] += round(loss_sparse.detach().cpu().item(), 5)
                eval_log['loss_ma'] += round(loss_margin.detach().cpu().item(), 5)
                eval_log['loss'] += lossE.detach().cpu().item()
                eval_log['lossG'] += lossG.detach().cpu().item()
                # eval_log['lossG'] += 0


        for log_item in eval_log.keys():
            # eval_log[log_item] /= (i+1)#lry change
            eval_log[log_item] = round(eval_log[log_item] / (i + 1), 5)

        # go through val samples
        if evaluation_turn:
            adj_matrix = adj_m.cpu().detach().numpy()
            adjs.append(adj_matrix)
            eval_log['negative_adj'] = int(np.sum(adj_matrix < -1e-5))

            # adjs.append(adj_m2)
            if ground_truth is not None:
                epoch_perf = get_metrics_auc(adj_matrix, ground_truth)
                for k in epoch_perf.keys():
                    eval_log[k] = epoch_perf[k]

            if has_train_test_split:
                eval_log['val_loss_rec'] = 0
                eval_log['val_loss_kl'] = 0
                eval_log['val_loss_sparse'] = 0
                vae.train(False)
                for batch in val_loader:
                    x = batch[0]
                    if configs['cuda']:
                        x = x.cuda()
                    out = vae(x, global_mean, global_std)
                    eval_log['val_loss_rec'] += out['loss_rec'].detach().cpu().item()
                    eval_log['val_loss_kl'] += out['loss_kl'].detach().cpu().item()
                    eval_log['val_loss_sparse'] += out['loss_sparse'].detach().cpu().item()
                if epoch >= configs['delayed_steps_on_sparse']:
                    es_tracks.append(eval_log['val_loss_rec'])

            # logger.log(eval_log)

            print(eval_log)
            # early stopping
            if (es > 0) and (len(es_tracks) > (es + 2)):
                if min(es_tracks[(-es - 1):]) < min(es_tracks[(-es):]):
                    print('Early stopping triggered')
                    break

    rec_xs = []

    logger.finish()
    vae = vae.cpu()
    # vae.classifier_pos_weight = vae.classifier_pos_weight.cpu()
    return vae, adjs, rec_xs




import numpy as np
import pandas as pd
from sklearn.metrics import average_precision_score

# Modified from 
# https://github.com/HantaoShu/DeepSEM/blob/master/src/utils.py
#https://github.com/TuftsBCB/dazzle/blob/main/dazzle/evaluate.py
from sklearn.metrics import average_precision_score,roc_auc_score
def get_metrics_auc(A, ground_truth):
    ''' Calculate Metrics including AUPR, AUC,AUPRR, EP, and EPR

    Calculate EPR given predicted adjacency matrix and BEELINE
    ground truth

    Parameters
    ----------
    A: numpy.array
        Predicted adjacency matrix. Expected size is |g| x |g|.
    ground_truth: tuple
        BEELINE ground truth object exported by
        data.load_beeline_ground_truth. The first element of this
        tuple is eval_flat_mask, the boolean mask on the flatten
        adjacency matrix to identify TFs and target genes. The
        second element is the lable values y_true after flatten.

    Returns
    -------
    tuple
        A tuple with AUPR,AUC, AUPR ratio, EP (in counts), and EPR
    '''
    eval_flat_mask, y_true, _ = ground_truth
    y_pred = np.abs(A.flatten()[eval_flat_mask])

    AUPR = average_precision_score(y_true, y_pred)
    AUC=roc_auc_score(y_true,y_pred)
    AUPRR = AUPR / np.mean(y_true)

    num_truth_edge = int(y_true.sum())
    cutoff = np.partition(y_pred, -num_truth_edge)[-num_truth_edge]
    y_above_cutoff = y_pred > cutoff
    EP = int(np.sum(y_true[y_above_cutoff]))
    EPR = 1. * EP / ((num_truth_edge ** 2) / np.sum(eval_flat_mask))
    # lry look edge
    return {'AUPR': AUPR,'AUROC':AUC, 'AUPRR': AUPRR,
            'EP': EP, 'EPR': EPR, 'true_edge': num_truth_edge}

def extract_edges(A, gene_names=None, TFmask=None, threshold=0.0):
    '''Extract predicted edges
    
    Extract edges from the predicted adjacency matrix
    
    Parameters
    ----------
    A: numpy.array 
        Predicted adjacency matrix. Expected size is |g| x |g|.
    gene_names: None, list or numpy.array
        (Optional) List of Gene Names. Usually accessible in the var_names 
        field of scanpy data. 
    TFmask: numpy.array
        A masking matrix indicating the position of TFs. Expected 
        size is |g| x |g|.
        
    Returns
    -------
    pandas.DataFrame
        A DataFrame including all the predicted links with predicted
        link strength.
    '''
    num_nodes = A.shape[0]
    mat_indicator_all = np.zeros([num_nodes, num_nodes])
    if TFmask is not None:
        A_masked = A * TFmask
    else:
        A_masked = A
    mat_indicator_all[abs(A_masked) > threshold] = 1
    idx_source, idx_target = np.where(mat_indicator_all)
    if gene_names is None:
        source_lbl = idx_source
        target_lbl = idx_target
    else:
        source_lbl = gene_names[idx_source]
        target_lbl = gene_names[idx_target]
    edges_df = pd.DataFrame(
        {'Source': source_lbl, 'Target': target_lbl, 
         'EdgeWeight': (A[idx_source, idx_target]),
         'AbsEdgeWeight': (np.abs(A[idx_source, idx_target]))
        })
    edges_df = edges_df.sort_values('AbsEdgeWeight', ascending=False)

    return edges_df.reset_index(drop=True)



import numpy as np
import scanpy as sc
import pandas as pd
import urllib
from tqdm import tqdm
import zipfile
import os

# Read ground truth
def load_beeline_ground_truth(data_dir, gene_names):
    n_gene = len(gene_names) # Get the number of genes
    ground_truth = pd.read_csv(f'{data_dir}/label.csv')# Read the label file
    TF = set(ground_truth['Gene1'])# Set of transcription factors
    All_gene = set(ground_truth['Gene1']) | set(ground_truth['Gene2'])# Set of all genes
    # Initialize evaluation mask and transcription factor mask
    evaluate_mask = np.zeros([n_gene, n_gene])
    TF_mask = np.zeros([n_gene, n_gene])
    for i, item in enumerate(gene_names):
        for j, item2 in enumerate(gene_names):
            if i == j:
                continue
            if item in TF and item2 in All_gene:
                evaluate_mask[i, j] = 1
            if item in TF:
                TF_mask[i, j] = 1
    # Create a DataFrame for the ground truth
    truth_df = pd.DataFrame(np.zeros([n_gene, n_gene]), 
                            index=gene_names, columns=gene_names)
    for i in range(ground_truth.shape[0]):
        truth_df.loc[ground_truth.iloc[i, 0], ground_truth.iloc[i, 1]] = 1
    A_truth = truth_df.values
    # Get indices of true edges
    idx_source, idx_target = np.where(A_truth)
    truth_edges = set(zip(idx_source, idx_target))
    
    eval_flat_mask = (evaluate_mask.flatten() != 0)# Flatten evaluation mask
    y_true = A_truth.flatten()[eval_flat_mask]
    
    return eval_flat_mask, y_true, truth_edges # Return evaluation mask, true labels, and true edges

def load_beeline(data_dir, benchmark_data='hESC', 
                 benchmark_setting='500_STRING'):
    ''' Load BEELINE
    
    Load BEELINE data into memory (download if necessary).
    
    Parameters
    ----------
    data_dir: str
        Root folder where the BEELINE data is/will be located. 
    benchmark_data: str
        Benchmark datasets. Choose among `hESC`, `hHep`, `mDC`, 
        `mESC`, `mHSC`, `mHSC-GM`, and `mHSC-L`.
    benchmark_setting: str
        Benchmark settings. Choose among `500_STRING`, 
        `1000_STRING`, `500_Non-ChIP`, `1000_Non-ChIP`, 
        `500_ChIP-seq`, `1000_ChIP-seq`, `500_lofgof`,
        and `1000_lofgof`. If either of the `lofgof` settings
        is choosed, only `mESC` data is available. 
        
    Returns
    -------
    tuple
        First element is a scanpy data with cells on rows and 
        genes on columns. Second element is the corresponding 
        BEELINE ground truth data 
    '''
    # Check if data directory exists, create if not
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)
    if not os.path.exists(f'{data_dir}/BEELINE/'):
        download_beeline(data_dir)
    data_dir = f'{data_dir}/BEELINE/{benchmark_setting}_{benchmark_data}'
    data = sc.read(f'{data_dir}/data.csv')
    # We do need to transpose the data to have cells on rows and genes on columns
    data = data.transpose()
    ground_truth = load_beeline_ground_truth(data_dir, data.var_names)
    return data, ground_truth

def download_beeline(save_dir, remove_zip=True):
    if not os.path.exists(save_dir):
        raise Exception("save_dir does not exist")
    zip_path = os.path.join(save_dir, 'BEELINE.zip')
    download_file('https://bcb.cs.tufts.edu/GRN-VAE/BEELINE.zip', 
                  zip_path) # Download BEELINE data
    with zipfile.ZipFile(zip_path,"r") as zip_ref:
        for file in tqdm(desc='Extracting', iterable=zip_ref.namelist(), 
                         total=len(zip_ref.namelist())):
            zip_ref.extract(member=file, path=save_dir)
    if remove_zip:
        os.remove(zip_path)

# Modified from https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51
def download_file(url, file_path, chunk_size=1024):
    req = urllib.request.urlopen(url)
    with open(file_path, 'wb') as f, tqdm(
        desc=f'Downloading {file_path}', total=req.length, unit='iB', 
        unit_scale=True, unit_divisor=1024
    ) as bar:
        for _ in range(req.length // chunk_size + 1):
            chunk = req.read(chunk_size)
            if not chunk: break
            size = f.write(chunk)
            bar.update(size)



"""
Single-cell Gene Regulatory Networks Inference and Analytics
"""
from . import data

from .logger import LightLogger, load_logger
from .models import SIGRN
from .SIGRN import runSIGRN
from .evaluator import extract_edges, get_metrics_auc

__all__ = [ 'load_beeline', 'LightLogger', 'load_logger',
           'SIGRN', 'runSIGRN', 'extract_edges', 'get_metrics_auc']


import os
import json
import datetime
import pandas as pd
import numpy as np

class LightLogger:
    """
    
    """
    
    def __init__(self, result_dir='result_logs', log_date=None):
        if log_date is None:
            log_date = str(datetime.date.today())
        self.result_dir = result_dir
        self.log_dir = f'{result_dir}/{log_date}'
        if not os.path.exists(result_dir):
            os.mkdir(result_dir)
        if not os.path.exists(self.log_dir):
            os.mkdir(self.log_dir)
        self.configs = {}
        self.mem = {}
        self.current_log = None
        self.logging_vars = set()
    
    def set_configs(self, configs):
        self.configs = configs
    
    def start(self, note=None):
        if note is None:
            note = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f") + str(np.random.choice(100000))
        self.current_log = note
        self.mem[note] = {'time': note}
        for k in self.configs.keys():
            self.mem[note][k] = self.configs[k]
        self.mem[note]['current_step'] = -1
        self.mem[note]['log'] = {}
        return note
        
    
    def log(self, log_dict, step=None):
        if step is None:
            step = self.mem[self.current_log]['current_step'] + 1
        self.mem[self.current_log]['current_step'] = step
        self.mem[self.current_log]['log'][step] = {}
        for k in log_dict.keys():
            self.mem[self.current_log]['log'][step][k] = log_dict[k]
            self.logging_vars.add(k)
            
    def finish(self, save_now=True):
        if save_now:
            with open(f'{self.log_dir}/{self.current_log}.json', 'w') as f:
                json.dump(self.mem[self.current_log], f)
        self.current_log = None
    
    def to_df(self, tidy=True):
        export_df = pd.DataFrame(self.mem).transpose().reset_index()
        if tidy:
            export_df['steps'] = export_df['log'].map(lambda x: list(x.keys()))
            for v in self.logging_vars:
                export_df[v] = export_df['log'].map(
                    lambda x: [x[k].get(v) for k in x.keys()]
                )
            del export_df['log']
            export_df = export_df.explode(
                ['steps'] + list(self.logging_vars), ignore_index=True)
        return export_df
    
    def save(self, path):
        export = {}
        export['result_dir'] = self.result_dir
        export['log_dir'] = self.log_dir
        export['configs'] = self.configs
        export['mem'] = self.mem
        export['current_log'] = self.current_log
        export['logging_vars'] = list(self.logging_vars)
        with open(path, 'w') as f:
            json.dump(export, f)
            
    def delete_batch(self, batch_name, filter_field='experiment_name'):
        to_delete = []
        for k in self.mem.keys():
            if self.mem[k][filter_field] == batch_name:
                to_delete.append(k)
        for k in to_delete:
            del self.mem[k]

def load_logger(path):
    with open(path, 'r') as f:
        logger_import = json.load(f)
    log_date = logger_import['log_dir'].replace(logger_import['result_dir']+'/', '')
    logger = LightLogger(logger_import['result_dir'], log_date=log_date)
    logger.set_configs(logger_import['configs'])
    logger.mem = logger_import['mem']
    logger.current_log = logger_import['current_log']
    logger.logging_vars = set(logger_import['logging_vars'])
    print("Loading logger complete!")
    return logger




import numpy as np
from .data import load_beeline
from .logger import LightLogger
from .SIGRN import runSIGRN
from .evaluate import extract_edges, get_metrics_auc
import time
#Model Configurations
# The three key concepts proposed in the SIGRN paper are controlled by the following parameters.
# delayed_steps_on_sparse: Number of delayed steps on introducing the sparse loss.
# dropout_augmentation: The proportion of data that will be randomly masked as dropout in each traing step.
# train_on_non_zero: Whether to train the model on non-zero expression data

DEFAULT_GRNVAE_CONFIGS = {
    # Train/Test split
    'train_split': 1.0,# Use all data for training
    'train_split_seed': None, # Seed for random splitting

    # Neural Net Definition
    'hidden_dim': 128,#  Size of dimension in the MLP layers
    'z_dim': 1,# Size of dimension of Z
    'A_dim': 1,#Number of Adjacency matrix to be modeled at the same time
    'train_on_non_zero': True,# Whether to train on non-zero data only
    'dropout_augmentation_p': 0.1,#Probability of augmented dropout. For example, 0.1 means that 10% of data will be temporarily assign to zero in each forward  pass
    'dropout_augmentation_type': 'all',#Choose among 'all' (default), 'belowmean', 'belowhalfmean'. This option specifies where dropout augmentation would happen. If 'belowmean' is selected, the augmentation would only happen on values below global mean.
    'cuda': True,

    # Loss term, hyperparameters
    'alpha': 100,
    # 'beta': 1,
    'chi': 0.5,
    'h_scale': 0,
    'delayed_steps_on_sparse': 30,

    # Neural Net Training
    'number_of_opt': 2,# Number of optimizations
    'batch_size': 64, # Size of training batches
    'n_epochs': 120, # Number of training epochs
    # 'schedule': [120, 240],
    'eval_on_n_steps': 10,# Evaluation frequency
    'early_stopping': 0,# Early stopping criteria
    'lr_nn': 1e-4, # Learning rate for neural network
    'lr_adj': 2e-5, # Learning rate for adjacency matrix
    'K1': 1,
    'K2': 1
}
# Load data from a BEELINE benchmark
# BEELINE benchmarks could be loaded by the load_beeline function, where you specify where to look for data and which benchmark to load. If it's the first time, this function will download the files automatically.
data, ground_truth = load_beeline(
    data_dir='data',
    benchmark_data='hESC', #hESC,hHep,mDC,mESC,mHSC-E,mHSC-GM,mHSC-L
    benchmark_setting='500_STRING'#500_STRING,1000_STRING,1000_Non-ChIP,500_Non-ChIP
)
print(data)
# Initialize logger to track training progress
logger = LightLogger()
start_time = time.time()

#Model Training
vae, adjs, result_rec = runSIGRN(
    data.X, DEFAULT_GRNVAE_CONFIGS, ground_truth=ground_truth, logger=logger)

# Output results
A = vae.get_adj()
end_time = time.time()
total_time = end_time - start_time  # seconds
#
ppi_auc = get_metrics_auc(A, ground_truth)
print(ppi_auc)




import torch
from torch import nn
from .nearest_embed import NearestEmbed,NearestEmbedEMA
class MLP(nn.Module):
    def __init__(self, in_dim, hidden_dim, out_dim, activation):
        super(MLP, self).__init__()
        # Define three fully connected layers
        self.l1 = nn.Linear(in_dim, hidden_dim)
        self.l2 = nn.Linear(hidden_dim, hidden_dim)
        self.l3 = nn.Linear(hidden_dim, out_dim)
        self.act = activation()

    def forward(self, x):
        # Forward pass through the MLP
        out1 = self.act(self.l1(x))
        out2 = self.act(self.l2(out1))
        return self.l3(out2)


class SIGRN(nn.Module):
    ''' A SIGRN model

    Parameters
    ----------
    n_genes: int
        Number of Genes
    hidden_dim: int
        Size of dimension in the MLP layers
    z_dim: int
        Size of dimension of Z
    A_dim: int
        Number of Adjacency matrix to be modeled at the same time
    activation: function
        A pytorch activation layer
    train_on_non_zero: bool
        Whether to train on non-zero data only
    dropout_augmentation_p: double
        Probability of augmented dropout. For example, 0.1 means that
        10% of data will be temporarily assign to zero in each forward
        pass
    dropout_augmentation_type: str
        Choose among 'all' (default), 'belowmean', 'belowhalfmean'. This
        option specifies where dropout augmentation would happen. If
        'belowmean' is selected, the augmentation would only happen on
        values below global mean.
    pretrained_A: torch.tensor
        A customized initialization of A instead of random initialization.

    Methods
    -------
    get_adj_
        Obtain current adjacency matrix
    get_adj
        Obtain current adjacency matrix as a detached numpy array
    I_minus_A
        Calculate I - A
    reparameterization(z_mu, z_sigma)
        Reparameterization trick used in VAE
    add_gaussian_noise(mean, std)
        Add gaussian noise to the original expression data
    dropout_augmentation2(x, global_mean)
        Randomly add dropout noise to the original expression data
    forward(x, global_mean, global_std, normal,add_gaussian,use_dropout_augmentation)
        Forward pass
    '''
    def __init__(
            self, n_gene, hidden_dim=128, z_dim=1, A_dim=1,
            activation=nn.Tanh, train_on_non_zero=False,
            dropout_augmentation_p=0.1, dropout_augmentation_type='all',
            pretrained_A=None,
    ):
        super(SIGRN, self).__init__()
        self.n_gene = n_gene
        self.hidden_dim = hidden_dim
        self.z_dim = z_dim
        self.A_dim = A_dim
        self.train_on_non_zero = train_on_non_zero

        # Initialize adjacency matrix A
        if pretrained_A is None:
            adj_A = torch.ones(A_dim, n_gene, n_gene) / (n_gene - 1)
            adj_A += torch.rand_like(adj_A) * 0.0002
        else:
            adj_A = pretrained_A
        self.adj_A = nn.Parameter(adj_A, requires_grad=True)

        # Inference and generative models
        self.inference_zposterior = MLP(1, hidden_dim, z_dim * 2, activation)
        self.generative_pxz = MLP(z_dim, hidden_dim, 1, activation)
        self.da_p = dropout_augmentation_p
        self.da_type = dropout_augmentation_type

        # Initialize weights
        self.apply(self._init_weights)

    def _init_weights(self, module):
        # Initialize weights using Xavier normal distribution
        if isinstance(module, nn.Linear):
            torch.nn.init.xavier_normal_(module.weight)
            if module.bias is not None:
                module.bias.data.zero_()

    def get_adj_(self):
        eye_tensor = torch.eye(
            self.n_gene, device=self.adj_A.device
        ).repeat(self.A_dim, 1, 1)
        mask = torch.ones_like(self.adj_A) - eye_tensor
        return (self.adj_A * mask).mean(0)

    def get_adj(self):
        # Return adjacency matrix as a numpy array
        return self.get_adj_().cpu().detach().numpy()

    def I_minus_A(self):
        # Calculate I - A
        eye_tensor = torch.eye(
            self.n_gene, device=self.adj_A.device
        ).repeat(self.A_dim, 1, 1)
        # clean up A along diagnal line
        mask = torch.ones_like(self.adj_A) - eye_tensor
        clean_A = self.adj_A * mask
        return eye_tensor - clean_A

    def reparameterization(self, z_mu, z_sigma):
        # Reparameterization trick for VAE
        return z_mu + z_sigma * torch.randn_like(z_sigma)

    @torch.no_grad()

    def add_gaussian_noise(self,matrix, mean=0, std=1):
        # Add Gaussian noise to the matrix
        device=matrix.device
        noise = torch.randn(matrix.shape,device=device) * std + mean
        noisy_matrix = matrix + noise
        return noisy_matrix
    def dropout_augmentation2(self, x, global_mean,da_p):
        # Apply mask augmentation
        da_mask = (torch.rand_like(x) < da_p)
        if self.da_type == 'belowmean':
            da_mask = da_mask * (x < global_mean)
        elif self.da_type == 'belowhalfmean':
            da_mask = da_mask * (x < (global_mean / 2))
        elif self.da_type == 'all':
            da_mask = da_mask
        noise =  x * da_mask
        x = x - noise
        return x, noise, da_mask
    def forward(self, x, global_mean,global_std, normal='z-score',add_gaussian=False,use_dropout_augmentation=True):
        if self.train_on_non_zero:
            eval_mask = (x != 0)
        else:
            eval_mask = torch.ones_like(x)

        x_init = x
        if use_dropout_augmentation:
            x, noise, da_mask = self.dropout_augmentation2(x_init, global_mean,self.da_p)

        else:
            noise = torch.zeros_like(x)
            da_mask = (noise == 1)

        if normal=='z-score':
            x = (x - global_mean) / (global_std)
            # noise = (noise - global_mean) / (global_std)
            x[torch.isnan(x)]=0
            x[torch.isinf(x)]=0

        else:
            x=x
            # noise=noise

        # Encoder --------------------------------------------------------------
        I_minus_A = self.I_minus_A()

        z_posterior = self.inference_zposterior(x.unsqueeze(-1))
        z_posterior = torch.einsum('ogd,agh->ohd', z_posterior, I_minus_A)

        z_mu = z_posterior[:, :, :self.z_dim]
        z_logvar = z_posterior[:, :, self.z_dim:]
        z = self.reparameterization(z_mu, torch.exp(z_logvar * 0.5))

        # Decoder --------------------------------------------------------------
        z_inv = torch.einsum('ogd,agh->ohd', z, torch.inverse(I_minus_A))
        x_rec = self.generative_pxz(z_inv).squeeze(2)

        # Losses ---------------------------------------------------------------
        loss_rec_all = (x - x_rec).pow(2)
        loss_rec = torch.sum(loss_rec_all * eval_mask)
        loss_rec = loss_rec / torch.sum(eval_mask)

        loss_kl = -0.5 * torch.mean(1 + z_logvar - z_mu.pow(2) - torch.exp(z_logvar))

        out = {
            'loss_rec': loss_rec, 'loss_kl': loss_kl,
            'z_posterior': z_posterior, 'z': z, 'x_rec': x_rec,
             'da_mask': da_mask,
            'norm_x':x,'z_inv':z_inv,'z_mu':z_mu,'IA':I_minus_A,'z_logvar':z_logvar
        }
        return out



import numpy as np
import torch
from torch import nn
from torch.autograd import Function, Variable
import torch.nn.functional as F


class NearestEmbedFunc(Function):
    """
    Nearest embedding function that finds the nearest embedding vector for each input vector.
    Input:
    ------
    x - (batch_size, emb_dim, *)
        Last dimensions may be arbitrary
    emb - (emb_dim, num_emb)
    """
    @staticmethod
    def forward(ctx, input, emb):
        if input.size(1) != emb.size(0):
            raise RuntimeError('invalid argument: input.size(1) ({}) must be equal to emb.size(0) ({})'.
                               format(input.size(1), emb.size(0)))

        # save sizes for backward
        ctx.batch_size = input.size(0)
        ctx.num_latents = int(np.prod(np.array(input.size()[2:])))
        ctx.emb_dim = emb.size(0)
        ctx.num_emb = emb.size(1)
        ctx.input_type = type(input)
        ctx.dims = list(range(len(input.size())))

        # expand to be broadcast-able
        x_expanded = input.unsqueeze(-1)
        num_arbitrary_dims = len(ctx.dims) - 2
        if num_arbitrary_dims:
            emb_expanded = emb.view(
                emb.shape[0], *([1] * num_arbitrary_dims), emb.shape[1])
        else:
            emb_expanded = emb

        # find nearest neighbors
        dist = torch.norm(x_expanded - emb_expanded, 2, 1)
        _, argmin = dist.min(-1)
        shifted_shape = [input.shape[0], *
                         list(input.shape[2:]), input.shape[1]]
        result = emb.t().index_select(0, argmin.view(-1)
                                      ).view(shifted_shape).permute(0, ctx.dims[-1], *ctx.dims[1:-1])

        ctx.save_for_backward(argmin)
        return result.contiguous(), argmin

    @staticmethod
    def backward(ctx, grad_output, argmin=None):
        grad_input = grad_emb = None
        if ctx.needs_input_grad[0]:
            grad_input = grad_output

        if ctx.needs_input_grad[1]:
            argmin, = ctx.saved_variables
            latent_indices = torch.arange(ctx.num_emb).type_as(argmin)
            idx_choices = (argmin.view(-1, 1) ==
                           latent_indices.view(1, -1)).type_as(grad_output.data)
            n_idx_choice = idx_choices.sum(0)
            n_idx_choice[n_idx_choice == 0] = 1
            idx_avg_choices = idx_choices / n_idx_choice
            grad_output = grad_output.permute(0, *ctx.dims[2:], 1).contiguous()
            grad_output = grad_output.view(
                ctx.batch_size * ctx.num_latents, ctx.emb_dim)
            grad_emb = torch.sum(grad_output.data.view(-1, ctx.emb_dim, 1) *
                                 idx_avg_choices.view(-1, 1, ctx.num_emb), 0)
        return grad_input, grad_emb, None, None


def nearest_embed(x, emb):
    return NearestEmbedFunc().apply(x, emb)


class NearestEmbed(nn.Module):
    def __init__(self, num_embeddings, embeddings_dim):
        super(NearestEmbed, self).__init__()
        self.weight = nn.Parameter(torch.rand(embeddings_dim, num_embeddings))

    def forward(self, x, weight_sg=False):
        """Input:
        ---------
        x - (batch_size, emb_size, *)
        """
        return nearest_embed(x, self.weight.detach() if weight_sg else self.weight)


# adapted from https://github.com/rosinality/vq-vae-2-pytorch/blob/master/vqvae.py#L25
# that adapted from https://github.com/deepmind/sonnet


class NearestEmbedEMA(nn.Module):
    def __init__(self, n_emb, emb_dim, decay=0.99, eps=1e-5):
        super(NearestEmbedEMA, self).__init__()
        self.decay = decay
        self.eps = eps
        self.embeddings_dim = emb_dim
        self.n_emb = n_emb
        self.emb_dim = emb_dim
        embed = torch.rand(emb_dim, n_emb)
        self.register_buffer('weight', embed)
        self.register_buffer('cluster_size', torch.zeros(n_emb))
        self.register_buffer('embed_avg', embed.clone())

    def forward(self, x):
        """Input:
        ---------
        x - (batch_size, emb_size, *)
        """

        dims = list(range(len(x.size())))
        x_expanded = x.unsqueeze(-1)
        num_arbitrary_dims = len(dims) - 2
        if num_arbitrary_dims:
            emb_expanded = self.weight.view(
                self.emb_dim, *([1] * num_arbitrary_dims), self.n_emb)
        else:
            emb_expanded = self.weight

        # find nearest neighbors
        dist = torch.norm(x_expanded - emb_expanded, 2, 1)
        _, argmin = dist.min(-1)
        shifted_shape = [x.shape[0], *list(x.shape[2:]), x.shape[1]]
        result = self.weight.t().index_select(
            0, argmin.view(-1)).view(shifted_shape).permute(0, dims[-1], *dims[1:-1])

        if self.training:
            latent_indices = torch.arange(self.n_emb).type_as(argmin)
            emb_onehot = (argmin.view(-1, 1) ==
                          latent_indices.view(1, -1)).type_as(x.data)
            n_idx_choice = emb_onehot.sum(0)
            n_idx_choice[n_idx_choice == 0] = 1
            flatten = x.permute(
                1, 0, *dims[-2:]).contiguous().view(x.shape[1], -1)

            self.cluster_size.data.mul_(self.decay).add_(
                1 - self.decay, n_idx_choice
            )
            embed_sum = flatten @ emb_onehot
            self.embed_avg.data.mul_(self.decay).add_(
                1 - self.decay, embed_sum)

            n = self.cluster_size.sum()
            cluster_size = (
                (self.cluster_size + self.eps) /
                (n + self.n_emb * self.eps) * n
            )
            embed_normalized = self.embed_avg / cluster_size.unsqueeze(0)
            self.weight.data.copy_(embed_normalized)

        return result, argmin
